/*
   Copyright Universite de Versailles Saint-Quentin en Yvelines 2009
   AUTHORS: Sebastien Briais, Sid Touati

   This file is part of SIRA.
   
   SIRA is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.
   
   SIRA is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with SIRA.  If not, see
   <http://www.gnu.org/licenses/>.
*/
#ifndef __ILP_SOLVE_H
#define __ILP_SOLVE_H

/** GLPK stub */

#include <glpk.h>
#include <math.h>

#define ILP_SOLVER_LO (1<<0)
#define ILP_SOLVER_UP (1<<1)
#define ILP_SOLVER_DB (ILP_SOLVER_LO | ILP_SOLVER_UP)

#define ILP_SOLVER_TYPE glp_prob
#define ILP_SOLVER_COEF_TYPE double

static void ilp_solver_init() {
  glp_term_out(GLP_OFF);
}

static glp_prob *ilp_solver_create(const char *name, int min_flag, int cols) {
  glp_prob *prob = glp_create_prob();
  if(prob == NULL) {
    return NULL;
  }
  glp_set_prob_name(prob, name);
  if(min_flag) {
    glp_set_obj_dir(prob, GLP_MIN); // minimisation problem    
  } else {
    glp_set_obj_dir(prob, GLP_MAX); // maximisation problem    
  }
  glp_add_cols(prob, cols);
  int i;
  for(i = 1; i <= cols; i++) {
    glp_set_col_kind(prob, i, GLP_IV);
  }
  return prob;
}

static void ilp_solver_set_col_double(glp_prob *prob, int i) {
  glp_set_col_kind(prob, i, GLP_CV);
}

static void ilp_solver_delete(glp_prob *prob) {
  glp_delete_prob(prob);
}

static void ilp_solver_cleanup() {
  glp_free_env();
}

static void ilp_solver_set_col_name(glp_prob *prob, int i, const char *name) {
  if(name != NULL) {
    glp_set_col_name(prob, i, name);
  }
}

static void ilp_solver_set_col_bounds(glp_prob *prob, int i, int bnds, double lower, double upper) {
  if(bnds & ILP_SOLVER_LO) {
    if(bnds & ILP_SOLVER_UP) {
      glp_set_col_bnds(prob, i, GLP_DB, lower, upper);
    } else {
      glp_set_col_bnds(prob, i, GLP_LO, lower, upper);
    }
  } else if(bnds & ILP_SOLVER_UP) {
    glp_set_col_bnds(prob, i, GLP_UP, lower, upper);
  } else {
    glp_set_col_bnds(prob, i, GLP_FR, lower, upper);
  }
}

static void ilp_solver_set_obj_coef(glp_prob *prob, int i, double coef) {
  glp_set_obj_coef(prob, i, coef);
}

static int ilp_solver_add_row(glp_prob *prob, const char *name, int nb, int idx[], double coef[], int bnds, double bound) {
  int i = glp_add_rows(prob, 1);
  if(name != NULL) {
    glp_set_row_name(prob, i, name);
  }
  glp_set_mat_row(prob, i, nb, idx, coef);
  if(bnds & ILP_SOLVER_LO) {
    glp_set_row_bnds(prob, i, GLP_LO, bound, 0);
  } else if(bnds & ILP_SOLVER_UP) {
    glp_set_row_bnds(prob, i, GLP_UP, 0, bound);
  }
  return i;
}

static int ilp_solver_solve(glp_prob *prob) {
  glp_simplex(prob, NULL);
  if(glp_intopt(prob, NULL) == 0) {
    if(glp_get_status(prob) == GLP_OPT) {    
      return TRUE;
    } 
  }
  return FALSE;
}

static int ilp_solver_get_col_value_int(glp_prob *prob, int i) {
  return rint(glp_get_col_prim(prob, i));
}

static double ilp_solver_get_col_value_double(glp_prob *prob, int i) {
  return glp_get_col_prim(prob, i);
}

static double ilp_solver_get_obj_value(glp_prob *prob) {
  return glp_get_obj_val(prob);
}

#ifdef DEBUG
static void ilp_solver_write_lp(glp_prob *prob, const char *name) {
  glp_write_lp(prob, NULL, name);
}
#endif

#endif
