/*
   Copyright Universite de Versailles Saint-Quentin en Yvelines 2009
   AUTHORS: Sebastien Briais, Sid Touati

   This file is part of SIRA.
   
   SIRA is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.
   
   SIRA is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with SIRA.  If not, see
   <http://www.gnu.org/licenses/>.
*/
#include <SCEDA/hashset.h>

#include <stdio.h>

#include "reuse.h"
#include "util.h"
#include <GDD/ddg.h>

/** Creates a reuse edge description of given characteristics.

    @param[in] source = source vertex 
    @param[in] target = target vertex
    @param[in] lambda = distance 

    @return a reuse edge description */
SIRA_ReuseEdge *SIRA_reuse_edge_create(SCEDA_Vertex *source, SCEDA_Vertex *target, int lambda) {
  SIRA_ReuseEdge *e = safe_malloc(sizeof(SIRA_ReuseEdge));
  e->source = source;
  e->target = target;
  e->lambda = lambda;
  return e;
}

/** Delete a reuse edge description.

    @param[in] e = reuse edge description */
void SIRA_reuse_edge_delete(SIRA_ReuseEdge *e) {
  free(e);
}

/** Computes the cost of a list of reuse edges (sum of all the distances).

    @param[in] reuse_edges = list of reuse edges

    @return the cost */                                                   
int SIRA_reuse_decisions_get_cost(SCEDA_List *reuse_edges) {
  int cost = 0;
  SCEDA_ListIterator r_edges;
  SCEDA_list_iterator_init(reuse_edges, &r_edges);
  while(SCEDA_list_iterator_has_next(&r_edges)) {
    SIRA_ReuseEdge *re = SCEDA_list_iterator_next(&r_edges);
    cost += re->lambda;
  }
  SCEDA_list_iterator_cleanup(&r_edges);
  return cost;
}

/** Apply a set of reuse decisions to the corresponding DDG.
    
    @param[in] g = DDG
    @param[in] type = type
    @param[in] reuse_edges = reuse edges of given type
    @param[in] ual_flag = UAL semantics switch
    
    @return the list of added edges (could be deleted with GDD_delete_edges) */
SCEDA_List *SIRA_reuse_apply_decisions(SCEDA_Graph *g, const char *type, SCEDA_List *reuse_edges, int ual_flag) {
  SCEDA_List *antidep_edges = SCEDA_list_create(NULL);

  SCEDA_ListIterator r_edges;
  SCEDA_list_iterator_init(reuse_edges, &r_edges);
  while(SCEDA_list_iterator_has_next(&r_edges)) {
    SIRA_ReuseEdge *re = SCEDA_list_iterator_next(&r_edges);
    
    SCEDA_Vertex *vi = re->source;
    SCEDA_Vertex *vj = re->target;
    int mu_ij = re->lambda;
    
    SCEDA_HashSet *cons_vi = GDD_consumers_of_type(vi, type);
    
    if(!SCEDA_hashset_is_empty(cons_vi)) {
      SCEDA_HashSetIterator consumers;
      SCEDA_hashset_iterator_init(cons_vi, &consumers);
      while(SCEDA_hashset_iterator_has_next(&consumers)) {
	SCEDA_Vertex *consi = SCEDA_hashset_iterator_next(&consumers);
	
	int delta = GDD_vertex_delta_r(consi, type);
	if(ual_flag) {
	  delta++;
	} else {
	  delta -= GDD_vertex_delta_w(vj, type);
	}
	
	int lambda = mu_ij - GDD_max_distance_flowdep(vi, consi, type);
	
	if((consi != vj) || (lambda != 0)) {
	  GDD_EdgeLabel *le = GDD_edge_label_create(type, "antidep_reg", delta, lambda);
	  SCEDA_Edge *ae = SCEDA_graph_add_edge(g, consi, vj, le);
	  safe_call(SCEDA_list_add(antidep_edges, ae));
	}
      }
      SCEDA_hashset_iterator_cleanup(&consumers);
    } else {
      // int delta = GDD_vertex_delta_r(vi, type); FIXED BUG
      int delta = 1;
      if(ual_flag) {
	delta++;
      } else {
	delta -= GDD_vertex_delta_w(vj, type);
      }
      int lambda = mu_ij;
      if((vi != vj) || (lambda != 0)) {
	GDD_EdgeLabel *le = GDD_edge_label_create(type, "antidep_reg", delta, lambda);
	SCEDA_Edge *ae = SCEDA_graph_add_edge(g, vi, vj, le);
	safe_call(SCEDA_list_add(antidep_edges, ae));
      }
    }
    
    SCEDA_hashset_delete(cons_vi);
  }
  SCEDA_list_iterator_cleanup(&r_edges);

  return antidep_edges;
}

#if(0)
static SIRA_EdgeDesc *edge_desc_create(SCEDA_Vertex *source, SCEDA_Vertex *target, int latency, int distance, const char *type) {
  SIRA_EdgeDesc *x = safe_malloc(sizeof(SIRA_EdgeDesc));
  x->source = source;
  x->target = target;
  x->latency = latency;
  x->distance = distance;
  x->type = safe_strdup(type);
  return x;
}

static void edge_desc_delete(SIRA_EdgeDesc *x) {
  free(x->type);
  free(x);
}

/** Get the list of antidep edge descriptions that correspond to reuse decisions.

    @param[in] g = DDG
    @param[in] type = type
    @param[in] reuse_edges = reuse edges of given type
    @param[in] ual_flag = UAL semantics switch 

    @return the list of (antidep) edge descriptions */
SCEDA_List *SIRA_reuse_decisions_to_edges_desc(SCEDA_Graph *g, const char *type, SCEDA_List *reuse_edges, int ual_flag) {
  SCEDA_List *antidep_edges = SCEDA_list_create((SCEDA_delete_fun)edge_desc_delete);

  SCEDA_ListIterator r_edges;
  SCEDA_list_iterator_init(reuse_edges, &r_edges);
  while(SCEDA_list_iterator_has_next(&r_edges)) {
    SIRA_ReuseEdge *re = SCEDA_list_iterator_next(&r_edges);
    
    SCEDA_Vertex *vi = re->source;
    SCEDA_Vertex *vj = re->target;
    int mu_ij = re->lambda;
    
    SCEDA_HashSet *cons_vi = GDD_consumers_of_type(vi, type);
    
    if(!SCEDA_hashset_is_empty(cons_vi)) {
      SCEDA_HashSetIterator consumers;
      SCEDA_hashset_iterator_init(cons_vi, &consumers);
      while(SCEDA_hashset_iterator_has_next(&consumers)) {
	SCEDA_Vertex *consi = SCEDA_hashset_iterator_next(&consumers);
	
	int delta = GDD_vertex_delta_r(consi, type);
	if(ual_flag) {
	  delta++;
	} else {
	  delta -= GDD_vertex_delta_w(vj, type);
	}
	
	int lambda = mu_ij - GDD_max_distance_flowdep(vi, consi, type);
	
	if((consi != vj) || (lambda != 0)) {
	  SIRA_EdgeDesc *ae = edge_desc_create(consi, vj, delta, lambda, type);
	  safe_call(SCEDA_list_add(antidep_edges, ae));
	}
      }
      SCEDA_hashset_iterator_cleanup(&consumers);
    } else {
      // int delta = GDD_vertex_delta_r(vi, type); FIXED BUG
      int delta = 1;
      if(ual_flag) {
	delta++;
      } else {
	delta -= GDD_vertex_delta_w(vj, type);
      }
      int lambda = mu_ij;
      if((vi != vj) || (lambda != 0)) {
	SIRA_EdgeDesc *ae = edge_desc_create(vi, vj, delta, lambda, type);
	safe_call(SCEDA_list_add(antidep_edges, ae));
      }
    }
    
    SCEDA_hashset_delete(cons_vi);
  }
  SCEDA_list_iterator_cleanup(&r_edges);

  return antidep_edges;
}
#endif
