/*
   Copyright Universite de Versailles Saint-Quentin en Yvelines 2009
   AUTHORS: Sebastien Briais, Sid Touati

   This file is part of SIRA.
   
   SIRA is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.
   
   SIRA is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with SIRA.  If not, see
   <http://www.gnu.org/licenses/>.
*/
#ifndef __SIRA_H
#define __SIRA_H
/** \file sira.h
    \brief Heuristics for reducing register saturation */

#include <SCEDA/list.h>
#include <SCEDA/graph.h>
#include "loop.h"
#include "weightmap.h"
#include "regconstraints.h"
#include "param.h"

/** Try to reduce register saturation of an acyclic DDG.

    @param[in] ddg = acyclic DDG
    @param[in] weights = weights of the types
    @param[out] new_edges = list of new antidep edges added to the DDG
    in case of success.

    @return 0 in case of success or a negative code otherwise

    Note that the DDG is modified in case of success.
*/
int SIRA_try_reduce_acyclic_saturation(SCEDA_Graph *ddg, SIRA_WeightMap *weights, SCEDA_List **new_edges);

/** Try to reduce register saturation of a cyclic DDG (innermost loop).

    @param[in] loop = innermost loop (SIRA loop)
    @param[in] settings = SiraLina settings (might be modified)
    @param[in] constraints = register constraints
    @param[in] weights = weights of the types (types are treated conjointly)
    @param[out] new_edges = list of new antidep edges added to the DDG
    in case of success.

    @return 0 in case of success or a negative code otherwise

    Note that the loop as well as the underlying DDG is modified in
    case of success.
*/
int SIRA_try_reduce_cyclic_saturation(SIRA_Loop *loop, SIRA_Param *settings, SIRA_RegConstraintsMap *constraints, SIRA_WeightMap *weights, SCEDA_List **new_edges);

/** Try to reduce register saturation of a cyclic DDG (innermost loop).

    @param[in] loop = innermost loop (SIRA loop)
    @param[in] settings = SiraLina settings (might be modified)
    @param[in] constraints = register constraints
    @param[in] types = list of types (types are treated successively)
    @param[out] new_edges = list of new antidep edges added to the DDG
    in case of success.

    @return 0 in case of success or a negative code otherwise

    Note that the loop as well as the underlying DDG is modified in
    case of success. 
*/
int SIRA_try_reduce_cyclic_saturation_iteratively(SIRA_Loop *loop, SIRA_Param *settings, SIRA_RegConstraintsMap *constraints, SCEDA_List *types, SCEDA_List **new_edges);

/** \page sira Schedule Independent Register Allocation (SIRA) heuristics

    To get a full understanding of SIRA library implementation, please also see:
    \li \ref siraloop
    \li \ref reuse
    \li \ref solver
    \li \ref isolver
    \li \ref redrs

    SIRA is based on the following research article: \li Karine
    Deschinkel and Sid Touati. <em>Efficient Method for
    Periodic Task Scheduling with Storage Requirement
    Minimization.</em> Proceedings of 2nd Annual International
    Conference on Combinatorial Optimization and Applications (COCOA
    2008). Saint Johns, Newfoundland, Canada. August 2008, LNCS
    Springer.  

    \section sira_api API

    SIRA offers three powerful register saturation reduction heuristics.

    \code
    int SIRA_try_reduce_acyclic_saturation(SCEDA_Graph *ddg, SIRA_WeightMap *weights, SCEDA_List **new_edges);
    \endcode

    This heuristic tries to reduce register saturation of an acyclic
    DDG.

    The weight of each type is given in a SIRA_WeightMap
    structure. See \ref weight_map for more details.

    \code
    int SIRA_try_reduce_cyclic_saturation(SIRA_Loop *loop, SIRA_Param *settings, SIRA_RegConstraintsMap *constraints, SIRA_WeightMap *weights, SCEDA_List **new_edges);
    \endcode

    This heuristics tries to reduce register saturation of a DDG of an
    innermost loop.

    Register types are treated conjointly.

    SIRALINA behaviour can be modified thanks to the SIRA_Param
    structure. See \ref param for more details.

    The weight of each type is given in a SIRA_WeightMap structure. See \ref
    weight_map for more details.

    The register constraints are given in a SIRA_RegConstraintsMap
    structure. See \ref regconstraints_map for more details.

    \code
    int SIRA_try_reduce_cyclic_saturation_iteratively(SIRA_Loop *loop, SIRA_Param *settings, SIRA_RegConstraintsMap *constraints, SCEDA_List *register_types, SCEDA_List **new_edges);
    \endcode

    This heuristics tries to reduce register saturation of a DDG of an
    innermost loop.

    Register types are treated one after the other. The order is given
    by the order in which they appear in the list.

    SIRALINA behaviour can be modified thanks to the SIRA_Param
    structure. See \ref param for more details.

    The register constraints are given in a SIRA_RegConstraintsMap
    structure. See \ref regconstraints_map for more details.

    \section sira_example Example

    \include "sira/main.c"
 */
#endif
